set linesize 32000
set pagesize 1000
set echo on
set termout on
set trimspool on
set serveroutput on size 1000000

spool bindmismatch.log

-- this script is meant to show the effect of having an indexed varchar2
-- with an equality condition in an update statement
-- and a number bind variable passed into in through an execute immediate
-- index is suppressed by the type mismatch
-- solution is to explicitly convert the number type column to a varchar2

-- create two tables to hold the plans
-- plan1 - plan with no index caused by type mismatch
-- plan2 - plan with efficient range scan of index

drop table plan1;
drop table plan2;

create table plan1
(ICOLUMN NUMBER,
 PLAN_TABLE_OUTPUT VARCHAR2(300));

create table plan2
(ICOLUMN NUMBER,
 PLAN_TABLE_OUTPUT VARCHAR2(300));
 
-- test table - indexed on test_column which is unique identifier

drop table test;

create table test
(OWNER  VARCHAR2(30),
 TABLE_NAME VARCHAR2(30),
 TEST_COLUMN VARCHAR2(2000));
 
create index testi on test(TEST_COLUMN);

insert into test select owner,table_name,to_char(rownum) from dba_tables;
commit;
insert into test select owner,table_name,to_char(rownum) from test;
commit;
insert into test select owner,table_name,to_char(rownum) from test;
commit;
insert into test select owner,table_name,to_char(rownum) from test;
commit;
insert into test select owner,table_name,to_char(rownum) from test;
commit;
insert into test select owner,table_name,to_char(rownum) from test;
commit;
insert into test select owner,table_name,to_char(rownum) from test;
commit;
insert into test select owner,table_name,to_char(rownum) from test;
commit;
insert into test select owner,table_name,to_char(rownum) from test;
commit;
insert into test select owner,table_name,to_char(rownum) from test;
commit;

execute dbms_stats.gather_table_stats(NULL,'TEST');

set timing on

-- run update passing number bind variable in where clause condition 
-- with equality to a varchar2

declare
V_NUMBER number;
begin

V_NUMBER := 1;

execute immediate 'UPDATE TEST SET TABLE_NAME=''UPDATED'' WHERE TEST_COLUMN=:1' using V_NUMBER;

insert into plan1 select rownum,plan_table_output from table(dbms_xplan.display_cursor(null,null,'ALL'));

commit;

end;
/

-- same update but with number variable converted explicitly to char by to_char

declare
V_NUMBER number;
begin

V_NUMBER := 1;

execute immediate 'UPDATE TEST SET TABLE_NAME=''UPDATEDAGAIN'' WHERE TEST_COLUMN=:1' using to_char(V_NUMBER);

insert into plan2 select rownum,plan_table_output from table(dbms_xplan.display_cursor(null,null,'ALL'));

commit;

end;
/

set timing off

-- note that the second execution was much faster than the first due to the use of the index

-- show plan for first update

select PLAN_TABLE_OUTPUT
from plan1
order by ICOLUMN;

-- show plan for second update

select PLAN_TABLE_OUTPUT
from plan2
order by ICOLUMN;

spool off
